/***********************************************************************

GECCO, a platform for command and control gaming

Copyright (C) 2001 Henrik Brnhielm, Andreas Enblom, Niklas
Hallenfur, Karl Hasselstrm, Henrik Hgerstrm, Oskar Linde, Jing Fu
Zi and Jon slund

This program is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 2 of the License, or (at
your option) any later version.

This program is distributed in the hope that it will be useful, but
WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307,
USA.

***********************************************************************/

package gecco.client;
import java.awt.*;
import java.awt.image.*;

import java.util.*;

/**
 *  This class represents the map of a game. The map should be given an Image
 *  and a map size. The map has a coordinate system of its own, where the upper
 *  left corner is (0,0), and every pixel is an automaton. The coordinates are
 *  then given as integer numbers. <p>
 *
 *  Every point of the map can be set as either visible or invisible. When a
 *  point is invisible it is painted darker. To make a point invisible, call
 *  hidePoint(). To make a point visible, or to change the state of som point
 *  (automaton), call method updatePoint. <p>
 *
 *
 *
 * @author     Andreas Enblom
 * @author     Karl Hasselstrm
 * @created    June 27, 2001
 */
public class Map {
    /**
     *  The size of the map, in automatons
     */
    private Dimension size;
    /**
     *  The color values of the automatons, one automaton per pixel.
     */
    private BufferedImage map;
    /**
     *  One flag for every automaton, set when it is visible, unset when it is
     *  invisible.
     */
    private boolean[][] visibleAutomatons;


    /**
     *  Initilizes a new map.
     *
     * @param  size  The size of the map, in automatons.
     */
    public Map() {
	clear();
    }


    /**
     *  Return the size of the map, in automatons.
     *
     * @return    The map size in automatons.
     */
    public Dimension getSize() {
        return new Dimension(size);
    }


    /**
     *  Return the center of the map, in map coordinates.
     *
     * @return    The center of the map.
     */
    public Position getCenter() {
        return new Position(((double) size.width) / 2.0,
                ((double) size.height) / 2.0);
    }


    /**
     *  Gets the image of automaton colors.
     *
     * @return    The automaton image.
     */
    public Image getImage() {
        return map;
    }


    /**
     *  Clears the map.
     */
    public void clear() {
        setSize(new Dimension(0, 0));
    }


    /**
     *  Loads the given image as map. The map is set as invisible.
     *
     * @param  map  Description of Parameter
     */
    public void loadMap(BufferedImage map) {

        // If the new map is null, just clear everything.
        if (map == null) {
            clear();
            return;
        }

	// Adjust the size.
	setSize(new Dimension(map.getWidth(), map.getHeight()));

        // the map is initially invisible
        for (int i = 0; i < size.width; ++i) {
            for (int j = 0; j < size.height; ++j) {
                visibleAutomatons[i][j] = false;
            }
        }
        Graphics g = this.map.getGraphics();

        // darken the map; it is initially invisible
        g.drawImage(map, 0, 0, size.width, size.height, null);
        g.setColor(new Color(0, 0, 0, 64));
        g.fillRect(0, 0, size.width, size.height);
    }


    /**
     *  Makes the given point invisibile by making it darker. If the point is
     *  already set as invisible, nothing is done.
     *
     * @param  x  Description of Parameter
     * @param  y  Description of Parameter
     */
    public void hidePoint(int x, int y) {
        if ((x >= size.width) || (y >= size.height) ||
                (x < 0) || (y < 0)) {
            return;
        }
        if (!visibleAutomatons[x][y]) {
            return;
        }
        visibleAutomatons[x][y] = false;

        // Make the color of the point 1/4 times darker (i.e.
        // new brightness = 3/4 * old brighness).
        int color = map.getRGB(x, y);
        map.setRGB(x, y, ((color >> 1) & 0x7F7F7F) +
                ((color >> 2) & 0x3F3F3F));
    }


    /**
     *  Updates the given point of the map to the new color. This marks the
     *  point as visible.
     *
     * @param  x      The x coordinate of the point.
     * @param  y      The y coordinate of the point.
     * @param  color  The color of the point, given as an int as described in
     *      java.awt.Color.
     */
    public void updatePoint(int x, int y, int color) {
        if ((x >= size.width) || (y >= size.height) ||
                (x < 0) || (y < 0)) {
            return;
        }
        visibleAutomatons[x][y] = true;
        map.setRGB(x, y, color);
    }

    
    public void setSize(Dimension size) {

	this.size = new Dimension(size);
	if (size.width > 0 || size.height > 0) {

	    // Make all automatons visible
	    visibleAutomatons = new boolean[size.width][size.height];
	    for (int i = 0; i < size.width; ++i) {
		for (int j = 0; j < size.height; ++j) {
		    visibleAutomatons[i][j] = true;
		}
	    }
	    
	    // Paint all automatons white
	    map = new BufferedImage(size.width, size.height,
				    BufferedImage.TYPE_3BYTE_BGR);
	    Graphics g = map.getGraphics();
	    g.setColor(Color.white);
	    g.fillRect(0, 0, size.width, size.height);

	} else {

	    map = null;
	}
    }

}
