/***********************************************************************

GECCO, a platform for command and control gaming

Copyright (C) 2001 Henrik Brnhielm, Andreas Enblom, Niklas
Hallenfur, Karl Hasselstrm, Henrik Hgerstrm, Oskar Linde, Jing Fu
Zi and Jon slund

This program is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 2 of the License, or (at
your option) any later version.

This program is distributed in the hope that it will be useful, but
WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307,
USA.

***********************************************************************/

package gecco.communication.client;

import java.awt.Color;
import java.awt.Point;
import java.io.*;
import java.util.*;

import gecco.client.*;
import gecco.communication.messages.*;

/**
 *  A class that listens for incoming messages from the server, and tells the
 *  client about them.
 *
 * @author     Niklas Hallenfur
 * @author     Karl Hasselstrm
 * @author     Henrik Hgerstrm
 * @author     Jon slund
 * @created    June 27, 2001
 */

public class DataInputExchangeHandler implements Runnable {

    /**
     *  Description of the Field
     */
    private ServerConnection connection;
    /**
     *  Description of the Field
     */
    private Client client;
    /**
     *  the ServerAmbassador who created this object
     */
    private ServerAmbassador i_am_working_for;
    /**
     *  set to true when the thread should end
     */
    private boolean die;
    /**
     *  only needed for Ping response
     */
    private DataOutputExchangeHandler DEHOutput;


    /**
     * @param  sc         connection to server containing the stream to write to
     * @param  cl         the Client to send incoming messages to
     * @param  sa         the ServerAmbassador this thread is working for
     *      DEHOutput the object to send PingReplyMessage:s to
     * @param  DEHOutput  Description of Parameter
     */
    public DataInputExchangeHandler(ServerConnection sc,
            Client cl, ServerAmbassador sa, DataOutputExchangeHandler DEHOutput) {
        connection = sc;
        client = cl;
        i_am_working_for = sa;
        die = false;
        this.DEHOutput = DEHOutput;
    }


    /**
     *  Main processing method for the DataInputExchangeHandler object
     */
    public void run() {
        while (!die) {
            try {
		Message m = Message.read(connection.getDataInputStream());
                handleMessage(m);
            } catch (Exception e) {

		// This is useful for debugging, but it makes the
		// client look broken when it's really not -- these
		// exceptions have been taken care of, so there's no
		// need to scare the user!
		//e.printStackTrace(System.out);

                i_am_working_for.connectionLost();
                break;
            }
            // Yield to other threads!
            Thread.yield();
        }
    }


    /**
     *  called whenever someone wants this thread to die
     */
    public void kill() {
        die = true;
    }


    /**
     *  A helper function that makes unsigned casting from byte to int
     *
     * @param  byteArray  Description of Parameter
     * @param  index      Description of Parameter
     * @return            The Int value
     */
    private int getInt(byte[] byteArray, int index) {
        // The given byte is always in the range 0-255, i.e. -128 - 127
        int foo = (int) byteArray[index];
        if (foo < 0) {
            return foo + 256;
        }
        return foo;
    }


    private void makeMapPointsInvisible(byte[] b) {

        int[] x = new int[b.length / 8];
        int[] y = new int[b.length / 8];

        for (int i = 0; i < b.length / 8; i++) {
            x[i] = (getInt(b, i * 8) << 24)
                     + (getInt(b, i * 8 + 1) << 16)
                     + (getInt(b, i * 8 + 2) << 8)
                     + getInt(b, i * 8 + 3);

            y[i] = (getInt(b, i * 8 + 4) << 24)
                     + (getInt(b, i * 8 + 5) << 16)
                     + (getInt(b, i * 8 + 6) << 8)
                     + getInt(b, i * 8 + 7);
        }

	client.hideMapPoints(x, y);
    }


    private void makeMapUpdates(byte[] b) {

        int[] x = new int[b.length / 11];
        int[] y = new int[b.length / 11];
        int[] c = new int[b.length / 11];

        for (int i = 0; i < b.length / 11; i++) {
            x[i] = (getInt(b, i * 11) << 24)
                     + (getInt(b, i * 11 + 1) << 16)
                     + (getInt(b, i * 11 + 2) << 8)
                     + getInt(b, i * 11 + 3);

            y[i] = (getInt(b, i * 11 + 4) << 24)
                     + (getInt(b, i * 11 + 5) << 16)
                     + (getInt(b, i * 11 + 6) << 8)
                     + getInt(b, i * 11 + 7);

            c[i] = (getInt(b, i * 11 + 8) << 16)
                     + (getInt(b, i * 11 + 9) << 8)
                     + getInt(b, i * 11 + 10);
       }

	client.mapUpdate(x, y, c);
    }


    /**
     *  Called when there are messages to receive from the server. Checks the
     *  type of message to get, and calls the appropriate method that gets that
     *  kind of message
     *
     * @param        m         Description of Parameter
     * @exception  Exception   Description of Exception
     */
    private void handleMessage(Message m) throws Exception {
	
	if (die) {
	    return;
	}

	if (m instanceof PingMessage) {
	    DEHOutput.addMessageToQueue(new PingReplyMessage());
	} else if (m instanceof UnitPositionUpdateMessage) {
	    UnitPositionUpdateMessage mm = (UnitPositionUpdateMessage) m;
            client.updatePiece(mm.handle, new Position(mm.x, mm.y));
	} else if (m instanceof UnitPropertyUpdateMessage) {
	    UnitPropertyUpdateMessage mm = (UnitPropertyUpdateMessage) m;
	    Property[] props = new Property[1];
	    props[0] = new Property(mm.key, mm.newValue);
            client.updatePiece(mm.handle, props);
	} else if (m instanceof UnitMessageMessage) {
	    UnitMessageMessage mm = (UnitMessageMessage) m;
            client.unitMessage(mm.handle, mm.message);
	} else if (m instanceof TimeUpdateMessage) {
	    TimeUpdateMessage mm = (TimeUpdateMessage) m;
            client.timeUpdate(mm.time);
	} else if (m instanceof UnitVisibleMessage) {
	    UnitVisibleMessage mm = (UnitVisibleMessage) m;
            client.addPiece(mm.handle, mm.name, new Position(mm.x, mm.y),
			    mm.unblockedActions, mm.properties, mm.type);
	} else if (m instanceof ActionBlockedMessage) {
	    ActionBlockedMessage mm = (ActionBlockedMessage) m;
	    if (mm.isBlocked) {
                client.removePieceAction(mm.handle, new Action(mm.name, mm.description,
							       mm.paramType, mm.instant));
	    } else {
                client.addPieceAction(mm.handle, new Action(mm.name, mm.description,
							    mm.paramType, mm.instant));
	    }
	} else if (m instanceof MapUpdateMessage) {
	    MapUpdateMessage mm = (MapUpdateMessage) m;
            client.mapUpdate(mm.x, mm.y, mm.color);
	} else if (m instanceof UnitDeletedMessage) {
	    UnitDeletedMessage mm = (UnitDeletedMessage) m;
            client.removePiece(mm.handle);
	} else if (m instanceof UnitDestroyedMessage) {
	    UnitDestroyedMessage mm = (UnitDestroyedMessage) m;
            client.destroyPiece(mm.handle);
	} else if (m instanceof MapPointInvisibleMessage) {
	    MapPointInvisibleMessage mm = (MapPointInvisibleMessage) m;
            client.hideMapPoint(mm.x, mm.y);
	} else if (m instanceof ActionCompletedMessage) {
	    ActionCompletedMessage mm = (ActionCompletedMessage) m;
            client.actionReply(mm.handle, true);
	} else if (m instanceof ActionFailedMessage) {
	    ActionFailedMessage mm = (ActionFailedMessage) m;
            client.actionReply(mm.handle, false);
	} else if (m instanceof MapPointsInvisibleMessage) {
	    MapPointsInvisibleMessage mm = (MapPointsInvisibleMessage) m;
	    makeMapPointsInvisible(mm.invisible);
	} else if (m instanceof MapUpdatesMessage) {
	    MapUpdatesMessage mm = (MapUpdatesMessage) m;
	    makeMapUpdates(mm.changes);
	} else {
	    throw new Exception("Wrong message!");
	}

    }

}
