/***********************************************************************

GECCO, a platform for command and control gaming

Copyright (C) 2001 Henrik Brnhielm, Andreas Enblom, Niklas
Hallenfur, Karl Hasselstrm, Henrik Hgerstrm, Oskar Linde, Jing Fu
Zi and Jon slund

This program is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 2 of the License, or (at
your option) any later version.

This program is distributed in the hope that it will be useful, but
WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307,
USA.

***********************************************************************/

package gecco.server.core;

import java.io.*;

import gecco.server.actionprocessor.ActionProcessor;
import gecco.server.automaton.AutomatonMap;
import gecco.server.clock.*;
import gecco.server.core.Client;
import gecco.server.core.Server;
import gecco.server.qmanager.QManager;
import gecco.server.unitmanager.UnitManager;
import gecco.server.visibility.*;

/**
 *  This class holds references to all important modules of the server. The
 *  modules can then get references to the other modules via this class.
 *
 * @author     Henrik Brnhielm
 * @author     Karl Hasselstrm
 * @author     Henrik Hgerstrm.
 * @created    April 30, 2001
 */
public class ReferenceHolder {
    /**
     *  Standard input
     */
    public static BufferedReader in;
    /**
     *  Standard output
     */
    public static PrintWriter out;
    /**
     *  Log output
     */
    public static PrintWriter log;
    /**
     *  Debug output
     */
    public static PrintWriter debug;
    /**
     *  Standard error output
     */
    public static PrintWriter err;

    /**
     *  Description of the Field
     */
    private static UnitManager unitManager;
    /**
     *  Description of the Field
     */
    private static QManager qManager;
    /**
     *  This is the visibility filter. The rest of the server tells it
     *  when something happens in the simulation, and it takes care of
     *  forwarding select pieces of this information to the clients.
     */
    private static VisibilityFilter visibilityFilter;
    /**
     *  This object can answer questions about who can see what.
     */
    private static VisibilityInfo visibilityInfo;
    /**
     *  Description of the Field
     */
    private static ActionProcessor actionProcessor;
    /**
     *  Description of the Field
     */
    private static AutomatonMap automatonMap;
    /**
     *  Description of the Field
     */
    private static Server comServer;
    /**
     *  This is the ClassLoader that is used to load classes from the jar file
     *  making up a particular game.
     */
    private static ClassLoader gameClassLoader;
    /**
     *  The clock that keeps track of the game time.
     */
    private static Clock clock;
    /**
     *  An object which can convert time to strings.
     */
    private static TimeString timeString;
    /**
     *  The name of the game.
     */
    private static String gameName;
    /**
     *  A short description of the game.
     */
    private static String gameDescription;


    /**
     *  Sets the AutomatonMap attribute of the ReferenceHolder class
     *
     * @param  map  The new AutomatonMap value
     */
    public static void setAutomatonMap(AutomatonMap map) {
        automatonMap = map;
    }


    /**
     *  Sets the ActionProcessor attribute of the ReferenceHolder class
     *
     * @param  _actionProcessor  The new ActionProcessor value
     */
    public static void setActionProcessor(ActionProcessor _actionProcessor) {
        actionProcessor = _actionProcessor;
    }


    /**
     *  Sets the UnitManager attribute of the ReferenceHolder class
     *
     * @param  _unitManager  The new UnitManager value
     */
    public static void setUnitManager(UnitManager _unitManager) {
        unitManager = _unitManager;
    }


    /**
     *  Sets the QManager attribute of the ReferenceHolder class
     *
     * @param  _qManager  The new QManager value
     */
    public static void setQManager(QManager _qManager) {
        qManager = _qManager;
    }


    /**
     *  Sets the ComServer attribute of the ReferenceHolder class
     *
     * @param  _comServer  The new ComServer value
     */
    public static void setComServer(Server _comServer) {
        comServer = _comServer;
    }


    /**
     *  Sets the visibility filter.
     *
     * @param  filter  The new visibility filter.
     */
    public static void setVisibilityFilter(VisibilityFilter filter) {
        visibilityFilter = filter;
    }


    /**
     *  Sets the visibility info object.
     *
     * @param  filter  The new visibility info object.
     */
    public static void setVisibilityInfo(VisibilityInfo info) {
        visibilityInfo = info;
    }


    /**
     *  Sets the clock.
     *
     * @param  clock  The new game clock.
     */
    public static void setClock(Clock clock) {
        ReferenceHolder.clock = clock;
    }


    /**
     *  Supplies an object that constructs strings representing time.
     *
     * @param timeString The new time string factory.
     */
    public static void setTimeString(TimeString timeString) {
        ReferenceHolder.timeString = timeString;
    }


    /**
     *  Sets the name of the game.
     *
     * @param name The new name.
     */
    public static void setGameName(String name) {
        ReferenceHolder.gameName = name;
    }


    /**
     *  Sets the description of the game.
     *
     * @param desc The new description.
     */
    public static void setGameDescription(String desc) {
        ReferenceHolder.gameDescription = desc;
    }


    /**
     *  Sets the GameClassLoader attribute of the ReferenceHolder class
     *
     * @param  loader  The new GameClassLoader value
     */
    public static void setGameClassLoader(ClassLoader loader) {
        gameClassLoader = loader;
    }


    /**
     *  Gets the AutomatonMap attribute of the ReferenceHolder class
     *
     * @return    The AutomatonMap value
     */
    public static AutomatonMap getAutomatonMap() {
        return automatonMap;
    }


    /**
     *  Gets the ActionProcessor attribute of the ReferenceHolder class
     *
     * @return    The ActionProcessor value
     */
    public static ActionProcessor getActionProcessor() {
        return actionProcessor;
    }


    /**
     *  Gets the UnitManager attribute of the ReferenceHolder class
     *
     * @return    The UnitManager value
     */
    public static UnitManager getUnitManager() {
        return unitManager;
    }


    /**
     *  Gets the QManager attribute of the ReferenceHolder class
     *
     * @return    The QManager value
     */
    public static QManager getQManager() {
        return qManager;
    }


    /**
     *  Gets the ComServer attribute of the ReferenceHolder class
     *
     * @return    The ComServer value
     */
    public static Server getComServer() {
        return comServer;
    }


    /**
     *  Gets the visibility filter.
     *
     * @return    The visibility filter.
     */
    public static VisibilityFilter getVisibilityFilter() {
        return visibilityFilter;
    }


    /**
     *  Gets the visibility info object.
     *
     * @return    The visibility info object.
     */
    public static VisibilityInfo getVisibilityInfo() {
        return visibilityInfo;
    }


    /**
     *  Gets the GameClassLoader attribute of the ReferenceHolder class
     *
     * @return    The GameClassLoader value
     */
    public static ClassLoader getGameClassLoader() {
        return gameClassLoader;
    }


    /**
     *  Gets the game clock.
     *
     * @return    The game clock.
     */
    public static Clock getClock() {
        return clock;
    }


    /**
     *  Gets the time string factory.
     *
     * @return    The time string factory.
     */
    public static TimeString getTimeString() {
        return timeString;
    }


    /**
     *  Gets the name of the game.
     *
     * @return The name of the game.
     */
    public static String getGameName() {
        return ReferenceHolder.gameName;
    }


    /**
     *  Gets a short description of the game.
     *
     * @return A short description of the game.
     */
    public static String getGameDescription() {
        return ReferenceHolder.gameDescription;
    }

}
