/***********************************************************************

GECCO, a platform for command and control gaming

Copyright (C) 2001 Henrik Brnhielm, Andreas Enblom, Niklas
Hallenfur, Karl Hasselstrm, Henrik Hgerstrm, Oskar Linde, Jing Fu
Zi and Jon slund

This program is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 2 of the License, or (at
your option) any later version.

This program is distributed in the hope that it will be useful, but
WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307,
USA.

***********************************************************************/

package gecco.server.startup;

import java.io.*;
import java.util.*;

import gecco.server.core.*;
import gecco.server.unitmanager.*;

/**
 *  This class loads the settings stored in unittypes.conf.
 *
 * @author     Henrik Brnhielm
 * @author     Karl Hasselstrm
 * @author     Henrik Hgerstrm
 * @created    July 11, 2001
 */
public abstract class UnitTypesConf {

    /**
     *  Parses the file and updates the unit manager.
     *
     * @param      in                    The configuration is read from this.
     * @param      um                    The unit manager.
     * @exception  IOException           If there is an I/O error.
     * @exception  SyntaxErrorException  If there is a syntax error.
     * @exception  ParseErrorException   If there is a parse error.
     */
    public static void parse(Reader in, UnitManager um)
             throws IOException, SyntaxErrorException, ParseErrorException {

        // Read the file into a data structure
        PropertyTree pt = new PropertyTree();
        pt.parseStream(in);

        // Extract information from the data structure
        Iterator iter = pt.getSubtreeList("unit_type").iterator();
        while (iter.hasNext()) {
            // Get the next unit type
            PropertyTree unit_type = (PropertyTree) iter.next();

            // Create the unit type
            UnitType currentUnitType = new
                    UnitType(unit_type.getSingleString("name"));

            // Some simple properties
            currentUnitType.setUnitClass(unit_type.getSingleString("class"));
            currentUnitType.setRange(unit_type.getSingleInt("range"));
            currentUnitType.setEventHandler(unit_type.getSingleString("event_handler"));

            // Load the icon
            String image = unit_type.getSingleString("image");
            currentUnitType.setImageData(StartServer.getImage(image));

            // Parse the actions
            Iterator iter2 = unit_type.getSubtreeList("action").iterator();
            while (iter2.hasNext()) {
                PropertyTree action = (PropertyTree) iter2.next();
                String name = action.getSingleString("name");
                String className = action.getSingleString("class");
                String argumentType = action.getSingleString("argument");
                String description = action.getSingleString("description");
                currentUnitType.addAction(name, description, argumentType, className);
            }

            // Parse the user-defined properties
            iter2 = unit_type.getSubtreeList("property").iterator();
            while (iter2.hasNext()) {
                PropertyTree property = (PropertyTree) iter2.next();
                String name = property.getSingleString("name");
                String type = property.getSingleString("type");

                // Make sure the property is not already defined as another type
                String oldType = currentUnitType.getPropertyType(name);
                if (oldType != null && !oldType.equals(type)) {
                    throw new ParseErrorException("Property " + name + " cannot be redefined to type " + type);
                }

                if (type.equals("STRING")) {
                    currentUnitType.setProperty(name, property.getSingleString("value"));
                } else if (type.equals("INT")) {
                    currentUnitType.setProperty(name, property.getSingleInt("value"));
                } else if (type.equals("DOUBLE")) {
                    currentUnitType.setProperty(name, property.getSingleDouble("value"));
                } else {
                    throw new ParseErrorException("Type must be STRING, INT or DOUBLE.");
                }
            }

            // Add the unit type
            um.addUnitType(currentUnitType);
        }
    }

}
