/***********************************************************************

GECCO, a platform for command and control gaming

Copyright (C) 2001 Henrik Brnhielm, Andreas Enblom, Niklas
Hallenfur, Karl Hasselstrm, Henrik Hgerstrm, Oskar Linde, Jing Fu
Zi and Jon slund

This program is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 2 of the License, or (at
your option) any later version.

This program is distributed in the hope that it will be useful, but
WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307,
USA.

***********************************************************************/

package gecco.server.visibility;

import gecco.server.actionprocessor.ActionProcessor;
import gecco.server.automaton.AutomatonMap;
import gecco.server.core.*;
import gecco.server.qmanager.QManager;
import gecco.server.unitmanager.UnitManager;

import java.awt.Color;
import java.util.*;

/**
 *  Description of the Class
 *
 * @author     Karl Hasselstrm
 * @author     Henrik Hgerstrm
 * @created    June 27, 2001
 */
public class VisibilityManager implements VisibilityFilter, VisibilityInfo {

    /**
     *  The part of the communication system that lets the server send
     *  messages to the clients.
     */
    private Client client;


    /**
     *  Sets the client (the part of the communication system that
     *  lets the server send messages to the clients).
     *
     * @param client The new client.
     */
    public void setClient(Client client) {
	this.client = client;
    }


    /**
     *  Description of the Method
     *
     * @param  role  Description of Parameter
     * @param  x     Description of Parameter
     * @param  y     Description of Parameter
     * @return       Description of the Returned Value
     */
    public boolean canSeePosition(String role, double x, double y) {
        Iterator it = ReferenceHolder.getUnitManager().getObservingUnits(role).iterator();
        while (it.hasNext()) {
            Unit u = (Unit) it.next();
            UnitType uT = u.getUnitType();
            double dx = u.getX() - x;
            double dy = u.getY() - y;
            double distSq = dx * dx + dy * dy;
            if (distSq <= uT.getRangeSquare()) {
                return true;
            }
        }
        return false;
    }


    /**
     *  Description of the Method
     *
     * @param  x  Description of Parameter
     * @param  y  Description of Parameter
     * @param  c  Description of Parameter
     */
    public void updateAutomatonState(int x, int y, int c) {
	String[] roles = ReferenceHolder.getUnitManager().getRoles();
        for (int i = 0; i < roles.length; i++) {
            if (ReferenceHolder.getUnitManager().isGodseye(roles[i])
		|| canSeeAutomaton(roles[i], x, y, null)) {
                client.automatonHasState(roles[i], x, y, c);
            }
        }
    }


    /**
     *  Description of the Method
     *
     * @param  role         Description of Parameter
     * @param  x            Description of Parameter
     * @param  y            Description of Parameter
     * @param  excludeUnit  Description of Parameter
     * @return              Description of the Returned Value
     */
    public boolean canSeeAutomaton(String role, int x, int y, Unit excludeUnit) {
        Iterator it = ReferenceHolder.getUnitManager().getObservingUnits(role).iterator();
        while (it.hasNext()) {
            Unit u = (Unit) it.next();
            if (u != excludeUnit && u.canSeeAutomatonAtPosition(x, y)) {
                return true;
            }
        }
        return false;
    }


    /**
     *  Returns, for every automaton on the map, true if the role can
     *  see it, and false otherwise.
     *
     * @param  role         The role.
     * @return A map of visibility info.
     */
    public boolean[][] visibleAutomatons(String role) {

	AutomatonMap am = ReferenceHolder.getAutomatonMap();
	UnitManager um = ReferenceHolder.getUnitManager();
	boolean[][] vis = new boolean[am.getWidth()][am.getHeight()];

	if (um.isGodseye(role)) {
	    for (int x = 0; x < am.getWidth(); x++) {
		for (int y = 0; y < am.getHeight(); y++) {
		    vis[x][y] = true;
		}
	    }
	} else {
	    // Just walk through all automatons and check if they are
	    // visible. This is probably the slowest, simplest way
	    // imaginable of doing this, but it is only done once when
	    // a client connects, so it's probably OK to be slow.
	    for (int x = 0; x < am.getWidth(); x++) {
		for (int y = 0; y < am.getHeight(); y++) {
		    vis[x][y] = canSeeAutomaton(role, x, y, null);
		}
	    }
	}

	return vis;
    }


    /**
     *  Description of the Method
     *
     * @param  role         Description of Parameter
     * @param  unit         Description of Parameter
     * @param  excludeUnit  Description of Parameter
     * @return              Description of the Returned Value
     */
    public boolean canSeeUnit(String role, Unit unit, Unit excludeUnit) {
        double x = unit.getX();
        double y = unit.getY();
        Iterator it = ReferenceHolder.getUnitManager().getObservingUnits(role).iterator();
        while (it.hasNext()) {
            Unit u = (Unit) it.next();
            if (u != excludeUnit) {
                UnitType uT = u.getUnitType();
                double dx = u.getX() - x;
                double dy = u.getY() - y;
                double distSq = dx * dx + dy * dy;
                if (distSq <= uT.getRangeSquare()) {
                    return true;
                }
            }
        }
        return false;
    }


    /**
     *  Returns the set of all units visible to the given role.
     *
     * @param  role         The role.
     * @return The units visible to that role.
     */
    public Set visibleUnits(String role) {
	
	UnitManager um = ReferenceHolder.getUnitManager();
	Collection allUnits = um.getAllUnits();
	Set visibleUnits;

	if (um.isGodseye(role)) {
	    visibleUnits = new HashSet(allUnits);
	} else {
	    visibleUnits = new HashSet();
	    Iterator iter = allUnits.iterator();
	    while (iter.hasNext()) {
		Unit u = (Unit) iter.next();
		if (canSeeUnit(role, u, null)) {
		    visibleUnits.add(u);
		}
	    }
	}

	return visibleUnits;
    }


    /**
     *  Description of the Method
     *
     * @param  unit  Description of Parameter
     * @param  oldX  Description of Parameter
     * @param  oldY  Description of Parameter
     * @param  newX  Description of Parameter
     * @param  newY  Description of Parameter
     */
    public void updateUnitPosition(Unit unit, double oldX, double oldY,
            double newX, double newY) {
	UnitManager um = ReferenceHolder.getUnitManager();
	AutomatonMap am = ReferenceHolder.getAutomatonMap();
        int handle = unit.getHandle();
        int range = um.getRange(unit);
        double rangeSq = range * range;
        String[] observingRoles = um.getRolesObservingUnit(unit);
        String[] alienRoles = um.getRolesAlienToUnit(unit);

        // Move unit for all godseye roles!
        Iterator godsEyeIt = um.getGodsEyeRoles().iterator();
        while (godsEyeIt.hasNext()) {
            String curRole = (String) godsEyeIt.next();
            client.unitMoved(curRole, handle, newX, newY);
        }

        // Send unit updates
        for (int i = 0; i < observingRoles.length; i++) {
            // First of all, send unit moved message since this role
            // is an observer, i.e. it should always be informed of
            // movements for this unit!
            String currentRole = observingRoles[i];
            client.unitMoved(currentRole, handle, newX, newY);

            // Fetch iterator for all alien units for this role!
            Iterator it = um.getAlienUnits(currentRole).iterator();
            while (it.hasNext()) {
                Unit alienUnit = (Unit) it.next();
                double x = (newX - alienUnit.getX());
                double y = (newY - alienUnit.getY());
                double distSq = x * x + y * y;
                if (distSq <= rangeSq) {
                    // Visible!
                    if (um.activateUnit(currentRole, alienUnit)) {
                        // The now visible unit wasn't visible before!
                        // Present new alien unit for role!
                        client.unitVisible(currentRole,
					   alienUnit.getHandle(),
					   alienUnit.getName(),
					   alienUnit.getUnitType().getName(),
					   alienUnit.getX(),
					   alienUnit.getY(),
					   null,
					   null,
					   null);
                    }
                } else {
                    // Not Visible!
                    if (!canSeeUnit(currentRole, alienUnit, unit) &&
                            um.deactivateUnit(currentRole, alienUnit)) {
                        // The now invisible unit was visible before,
                        // and this role cannot se it!
                        // Invisibilise alien unit for role!
                        client.unitInvisible(currentRole,
					     alienUnit.getHandle());
                    }
                }
            }
        }

        for (int i = 0; i < alienRoles.length; i++) {
            // Here we check if the moved unit is spotted
            // by the alien roles for this unit, i.e. the roles
            // that aren't observers or commander for the moved unit.
            String currentRole = alienRoles[i];

            if (canSeeUnit(currentRole, unit, null)) {
                if (um.activateUnit(currentRole, unit)) {
                    // The now visible unit wasn't visible before!
                    // Present the moved unit as an alien unit for the role!
                    client.unitVisible(currentRole,
				       unit.getHandle(),
				       unit.getName(),
				       unit.getUnitType().getName(),
				       unit.getX(),
				       unit.getY(),
				       null,
				       null,
				       null);
                } else {
                    // The moved unit was visible before! Just inform the role
                    // that it has moved!
                    client.unitMoved(currentRole, unit.getHandle(), newX, newY);
                }
            } else {
                if (um.deactivateUnit(currentRole, unit)) {
                    // The now invisible unit was visible before!
                    // Therefore, invisibilise the moved unit for role!
                    client.unitInvisible(currentRole,
					 unit.getHandle());
                }
            }
        }

        // Check for new visible/invisible automatons
        // Java's operation when casting a double to an int is to
        // floor the double, i.e. oldX = 1.94 -> oldIntX = 1
        int oldIntX = (int) oldX;
        int oldIntY = (int) oldY;
        int newIntX = (int) newX;
        int newIntY = (int) newY;
        UnitType uT = unit.getUnitType();
        int[] visionStart = uT.getVisionStart();
        int[] visionStop = uT.getVisionStop();
        int diff = oldIntY - newIntY;
        Iterator rolesIterator = unit.getObservers().iterator();
        while (rolesIterator.hasNext()) {
            String currentRole = (String) rolesIterator.next();
            HashSet units = new HashSet(um.getObservingUnits(currentRole));
            units.remove(unit);
            // Remove current unit from the set, since we always
            // want to check if any _other_ unit can see a specific
            // point or not.
            Unit[] unitsArray = (Unit[]) units.toArray(new Unit[0]);
            int curOldIndex = 0;
            int curNewIndex = 0;
            int curY;
            LinkedList visiblePoints = new LinkedList();
            LinkedList invisiblePoints = new LinkedList();
            if (diff < 0) {
                // Old circle on top
                curY = oldIntY - range;
                for (int i = 0; i < (-diff); i++, curY++, curOldIndex++) {
                    if (curY < 0) {
                        continue;
                    }
                    if (curY >= am.getHeight()) {
                        break;
                    }
                    // Iterate over all points in the interval, check if they are
                    // visible through some other unit. If not, add to invisible set.
                    int startX = visionStart[curOldIndex] + oldIntX;
                    int stopX = visionStop[curOldIndex] + oldIntX;
                    if (startX < 0) {
                        startX = 0;
                    }
                    if (stopX >= am.getWidth()) {
                        stopX = am.getWidth() - 1;
                    }
                    for (int k = startX; k <= stopX; k++) {
                        boolean found = false;
                        for (int l = 0; l < unitsArray.length; l++) {
                            if (unitsArray[l].canSeeAutomatonAtPosition(k, curY)) {
                                found = true;
                                break;
                            }
                        }
                        if (!found) {
                            invisiblePoints.add(new Point(k, curY));
                        }
                    }
                }
            } else {
                // New circle on top
                curY = newIntY - range;
                for (int i = 0; i < diff; i++, curY++, curNewIndex++) {
                    if (curY < 0) {
                        continue;
                    }
                    if (curY >= am.getHeight()) {
                        break;
                    }
                    // Iterate over all points in the interval, check if they are
                    // visible through some other unit. If not, add to visible set.
                    int startX = visionStart[curNewIndex] + newIntX;
                    int stopX = visionStop[curNewIndex] + newIntX;
                    if (startX < 0) {
                        startX = 0;
                    }
                    if (stopX >= am.getWidth()) {
                        stopX = am.getWidth() - 1;
                    }
                    for (int k = startX; k <= stopX; k++) {
                        boolean found = false;
                        for (int l = 0; l < unitsArray.length; l++) {
                            if (unitsArray[l].canSeeAutomatonAtPosition(k, curY)) {
                                found = true;
                                break;
                            }
                        }
                        if (!found) {
                            visiblePoints.add(new Point(k, curY));
                        }
                    }
                }
            }

            // INVARIANT: Both circles 'active' for next row (it's the first row for which
            // both circles are active).

            int checkValue = (2 * range + 1 - Math.abs(diff));
            for (int i = 0; i < checkValue; i++, curY++, curOldIndex++, curNewIndex++) {
                if (curY < 0) {
                    continue;
                }
                if (curY >= am.getHeight()) {
                    break;
                }

                // Kolla skrning mellan cirklarna, kontrollera om punkter r synliga
                // pga andra enheters synflt, och lgg till i ngon mngd om s behvs.
                int oldStartX = visionStart[curOldIndex] + oldIntX;
                int oldStopX = visionStop[curOldIndex] + oldIntX;
                int newStartX = visionStart[curNewIndex] + newIntX;
                int newStopX = visionStop[curNewIndex] + newIntX;
                if (oldStartX < 0) {
                    oldStartX = 0;
                }
                if (oldStopX >= am.getWidth()) {
                    oldStopX = am.getWidth() - 1;
                }
                if (newStartX < 0) {
                    newStartX = 0;
                }
                if (newStopX >= am.getWidth()) {
                    newStopX = am.getWidth() - 1;
                }

                // We have six cases (swap x and y). See below.

// 1)           xxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxx
//                               yyyyyyyyyyyyyyyyyyyyyyy

// 2)           xxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxx
//                    yyyyyyyyyyyyyyyyyyyyyyy

// 3)           xxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxx
//                                                      yyyyyyyyyyyyyyyyyyyyyyy

// The algorithm 'sorts' the four x-indeces, and sets the variables
// to the correct values.

                // Lefthand side interval
                LinkedList leftHandSideSet;
                LinkedList rightHandSideSet;
                int startLeft;
                int stopLeft;
                int startRight;
                int stopRight;

                // Spaghetti code...
                if (newStartX < oldStartX) {
                    // Sorted order so far: newStartX < oldStartX

                    if (oldStartX < newStopX) {
                        // Sorted order so far: newStartX < oldStartX < newStopX
                        leftHandSideSet = visiblePoints;
                        startLeft = newStartX;
                        stopLeft = oldStartX - 1;

                        if (newStopX < oldStopX) {

                            // Sorted order so far: newStartX < oldStartX < newStopX < oldStopX
                            // CASE 1 - new is 'x', old is 'y'
                            rightHandSideSet = invisiblePoints;
                            startRight = newStopX + 1;
                            stopRight = oldStopX;
                        } else {

                            // Sorted order so far: newStartX < oldStartX < oldStopX <= newStopX
                            // CASE 2 - new is 'x', old is 'y'
                            rightHandSideSet = visiblePoints;
                            startRight = oldStopX + 1;
                            stopRight = newStopX;
                        }
                    } else {
                        // Sorted order so far: newStartX <= newStopX <= oldStartX <= oldStopX
                        // CASE 3 - new is 'x', old is 'y'
                        leftHandSideSet = visiblePoints;
                        startLeft = newStartX;
                        stopLeft = newStopX;
                        rightHandSideSet = invisiblePoints;
                        startRight = oldStartX;
                        stopRight = oldStopX;
                        // If these two points coincide, we have to increase
                        // startRight by 1, and decrease stopLeft by 1
                        if (newStopX == oldStartX) {
                            stopLeft--;
                            startRight++;
                        }
                    }
                } else {
                    // Sorted order so far: oldStartX <= newStartX

                    if (newStartX < oldStopX) {
                        // Sorted order so far: oldStartX <= newStartX < oldStopX
                        leftHandSideSet = invisiblePoints;
                        startLeft = oldStartX;
                        stopLeft = newStartX - 1;

                        if (oldStopX < newStopX) {
                            // Sorted order so far: oldStartX <= newStartX < oldStopX < newStopX
                            // CASE 1 - old is 'x', new is 'y'
                            rightHandSideSet = visiblePoints;
                            startRight = oldStopX + 1;
                            stopRight = newStopX;

                        } else {
                            // Sorted order so far: oldStartX <= newStartX <= newStopX <= oldStopX
                            // CASE 2 - old is 'x', new is 'y'
                            rightHandSideSet = invisiblePoints;
                            startRight = newStopX + 1;
                            stopRight = oldStopX;
                        }
                    } else {
                        // Sorted order so far: oldStartX <= oldStopX <= newStartX <= newStopX
                        // CASE 3 - old is 'x', new is 'y'
                        leftHandSideSet = invisiblePoints;
                        startLeft = oldStartX;
                        stopLeft = oldStopX;
                        rightHandSideSet = visiblePoints;
                        startRight = newStartX;
                        stopRight = newStopX;
                        // If these two points coincide, we have to increase
                        // startRight by 1, and decrease stopLeft by 1
                        if (oldStopX == newStartX) {
                            stopLeft--;
                            startRight++;
                        }
                    }
                }

                for (int k = startLeft; k <= stopLeft; k++) {
                    boolean found = false;
                    for (int l = 0; l < unitsArray.length; l++) {
                        if (unitsArray[l].canSeeAutomatonAtPosition(k, curY)) {
                            found = true;
                            break;
                        }
                    }
                    if (!found) {
                        leftHandSideSet.add(new Point(k, curY));
                    }
                }

                // Righthand side interval
                for (int k = startRight; k <= stopRight; k++) {
                    boolean found = false;
                    for (int l = 0; l < unitsArray.length; l++) {
                        if (unitsArray[l].canSeeAutomatonAtPosition(k, curY)) {
                            found = true;
                            break;
                        }
                    }
                    if (!found) {
                        rightHandSideSet.add(new Point(k, curY));
                    }
                }

            }

            if (diff < 0) {
                // New circle below
                for (int i = 0; i < (-diff); i++, curY++, curNewIndex++) {
                    if (curY < 0) {
                        continue;
                    }
                    if (curY >= am.getHeight()) {
                        break;
                    }
                    // Iterate over all points in the interval, check if they are
                    // visible through some other unit. If not, add to visible set.
                    int startX = visionStart[curNewIndex] + newIntX;
                    int stopX = visionStop[curNewIndex] + newIntX;
                    if (startX < 0) {
                        startX = 0;
                    }
                    if (stopX >= am.getWidth()) {
                        stopX = am.getWidth() - 1;
                    }
                    for (int k = startX; k <= stopX; k++) {
                        boolean found = false;
                        for (int l = 0; l < unitsArray.length; l++) {
                            if (unitsArray[l].canSeeAutomatonAtPosition(k, curY)) {
                                found = true;
                                break;
                            }
                        }
                        if (!found) {
                            visiblePoints.add(new Point(k, curY));
                        }
                    }
                }
            } else {
                // Old circle below
                for (int i = 0; i < diff; i++, curOldIndex++, curY++) {
                    if (curY < 0) {
                        continue;
                    }
                    if (curY >= am.getHeight()) {
                        break;
                    }
                    // Iterate over all points in the interval, check if they are
                    // visible through some other unit. If not, add to invisible set.
                    int startX = visionStart[curOldIndex] + oldIntX;
                    int stopX = visionStop[curOldIndex] + oldIntX;
                    if (startX < 0) {
                        startX = 0;
                    }
                    if (stopX >= am.getWidth()) {
                        stopX = am.getWidth() - 1;
                    }
                    for (int k = startX; k <= stopX; k++) {
                        boolean found = false;
                        for (int l = 0; l < unitsArray.length; l++) {
                            if (unitsArray[l].canSeeAutomatonAtPosition(k, curY)) {
                                found = true;
                                break;
                            }
                        }
                        if (!found) {
                            invisiblePoints.add(new Point(k, curY));
                        }
                    }
                }
            }

            // Skicka meddelanden fr de tv mngderna med punkter...

            client.automatonsInvisible(currentRole, invisiblePoints);

            LinkedList visibleColors = new LinkedList();
            Iterator vis = visiblePoints.iterator();
            while (vis.hasNext()) {
                Point p = (Point) vis.next();
                visibleColors.add(am.getAutomaton(p.x, p.y).getCurrentColor());
            }
            client.automatonsHaveState(currentRole,
				      visiblePoints,
				      visibleColors);
        }

    }


    /**
     *  Description of the Method
     *
     * @param  unit        Description of Parameter
     * @param  lastKnownX  Description of Parameter
     * @param  lastKnownY  Description of Parameter
     */
    public void deleteUnit(Unit unit, double lastKnownX, double lastKnownY) {
        /*
         * For all observers of the unit, remove all automatons that aren't visible through
         * other observer-units.
         * Also remove all units that are visible through the unit.
         * Then report to all observer roles and all roles that has this unit as active
         * that it has been destroyed.
         *
         * You could think of this procedure as if the unit suddenly moves a long long
         * way out of our coordinate system.
         */

	UnitManager um = ReferenceHolder.getUnitManager();
	AutomatonMap am = ReferenceHolder.getAutomatonMap();
        int handle = unit.getHandle();
        int range = um.getRange(unit);
        double rangeSq = range * range;
        String[] observingRoles = um.getRolesObservingUnit(unit);
        String[] alienRoles = um.getRolesAlienToUnit(unit);

        // Delete unit for all godseye roles!
        Iterator godsEyeIt = um.getGodsEyeRoles().iterator();
        while (godsEyeIt.hasNext()) {
            String curRole = (String) godsEyeIt.next();
            client.unitDeleted(curRole, handle);
        }

        // Send unit updates
        for (int i = 0; i < observingRoles.length; i++) {
            // First of all, send unit destroyed message since this role
            // is an observer, i.e. it should always be informed of
            // movements for this unit!
            String currentRole = observingRoles[i];
            client.unitDeleted(currentRole, handle);

            LinkedList points = new LinkedList();

            // Fetch iterator for all alien units for this role!
            Iterator it = um.getAlienUnits(currentRole).iterator();
            while (it.hasNext()) {
                Unit alienUnit = (Unit) it.next();
                // Check if the role can't see the alien unit after
                // the argument unit has been destroyed
                if (!canSeeUnit(currentRole, alienUnit, unit) &&
                        um.deactivateUnit(currentRole, alienUnit)) {
                    // The now invisible unit was visible before,
                    // and this role cannot se it!
                    // Invisibilise alien unit for role!
                    client.unitInvisible(currentRole,
					 alienUnit.getHandle());
                }
            }

            // Now, iterate over all the automatons this unit could see
            // and store all the points that now has become invisible to the role.
            int x = (int) lastKnownX;
            int y = (int) lastKnownY;
            int[] xstart = unit.getUnitType().getVisionStart();
            int[] xstop = unit.getUnitType().getVisionStop();
            int curY = y - range;
            for (int k = 0; k < 2 * range + 1; k++, curY++) {
                if (curY < 0) {
                    continue;
                }
                if (curY >= am.getHeight()) {
                    break;
                }
                int left = xstart[k] + x;
                int right = xstop[k] + x;
                if (left < 0) {
                    left = 0;
                } else if (left >= am.getWidth()) {
                    continue;
                }
                if (right < 0) {
                    continue;
                } else if (right >= am.getWidth()) {
                    right = am.getWidth() - 1;
                }
                for (int j = left; j <= right; j++) {
                    if (!canSeeAutomaton(currentRole, j, curY, unit)) {
                        points.add(new Point(j, curY));
                    }
                }
            }
            // Invisibilise the automatons for this role!
            if (!points.isEmpty()) {
                client.automatonsInvisible(currentRole, points);
            }
        }

        // Lastly, send unit destroyed message to all roles that
        // are alien to the unit but saw it before it was destroyed!

        for (int i = 0; i < alienRoles.length; i++) {
            // Iterate over all alien roles and see if
            // the destroyed unit was active for any one of them.
            // If so, send a unit destroyed message.
            if (um.deactivateUnit(alienRoles[i], unit)) {
                client.unitDeleted(alienRoles[i], handle);
            }
        }
    }


    /**
     *  Description of the Method
     *
     * @param  role  Description of Parameter
     */
    public void resendAllData(String role) {
        // This is just a refresh of the current situation for a role!

        // SYNCHRONIZE ON THE QMANAGER QUEUES AND THE ACTIONPROCESSOR QUEUE,
        // OTHERWISE WE MIGHT HAVE SOME INTEGRITY PROBLEMS!
        LinkedList points = new LinkedList();
        LinkedList colors = new LinkedList();
        QManager qM = ReferenceHolder.getQManager();
        ActionProcessor aP = ReferenceHolder.getActionProcessor();
	UnitManager um = ReferenceHolder.getUnitManager();
	AutomatonMap am = ReferenceHolder.getAutomatonMap();
        synchronized (qM.getAutomatonQueue()) {
            synchronized (qM.getAutomatonEventQueue()) {
                synchronized (aP.getActionQueue()) {
                    // Ok, we've frozen all action going on in the game!
                    // Let's rock!

                    // If this role is a gods eye role, call startGodsEyeGame...
                    if (um.isGodseye(role)) {
                        resendAllGodsEyeData(role);
                        return;
                    }

                    // First, present all observer units and the automatons
                    // they can see!
                    Iterator it = um.getObservingUnits(role).iterator();
                    while (it.hasNext()) {
                        // Present the unit! Check if the role is the commander or just
                        // an observer!
                        Unit unit = (Unit) it.next();
                        UnitType unitType = unit.getUnitType();
                        if (role.equals(unit.getCommander())) {
                            Set unblockedActionDefinitions = new HashSet(unitType.getActionDefinitions());
                            Set blockedActionDefinitions = new HashSet();
                            Set blockedActionNames = unit.getBlockedActions();
                            Iterator actionIt = blockedActionNames.iterator();
                            while (actionIt.hasNext()) {
                                String name = (String) actionIt.next();
                                ActionDefinition actDef = unitType.getActionDefinition(name);
                                blockedActionDefinitions.add(actDef);
                                unblockedActionDefinitions.remove(actDef);
                            }
                            um.activateUnit(role, unit);
                            client.unitVisible(role,
					       unit.getHandle(),
					       unit.getName(),
					       unit.getUnitType().getName(),
					       unit.getX(),
					       unit.getY(),
					       unblockedActionDefinitions,
					       blockedActionDefinitions,
					       unit.getProperties());
                        } else {
                            // Role is an observer for the unit!
                            um.activateUnit(role, unit);
                            client.unitVisible(role,
					       unit.getHandle(),
					       unit.getName(),
					       unit.getUnitType().getName(),
					       unit.getX(),
					       unit.getY(),
					       null,
					       null,
					       unit.getProperties());
                        }
			
                        // Ok, now present all automatons this unit can see!
                        int x = unit.getIntX();
                        int y = unit.getIntY();
                        int range = unitType.getRange();
                        int[] xstart = unitType.getVisionStart();
                        int[] xstop = unitType.getVisionStop();
                        int curY = y - range;
                        for (int i = 0; i < 2 * range + 1; i++, curY++) {
                            if (curY < 0) {
                                continue;
                            }
                            if (curY >= am.getHeight()) {
                                break;
                            }
                            int left = xstart[i] + x;
                            int right = xstop[i] + x;
                            if (left < 0) {
                                left = 0;
                            } else if (left >= am.getWidth()) {
                                continue;
                            }
                            if (right < 0) {
                                continue;
                            } else if (right >= am.getWidth()) {
                                right = am.getWidth() - 1;
                            }
                            for (int j = left; j <= right; j++) {
                                points.add(new Point(j, curY));
                                colors.add(am.getAutomaton(j, curY).getCurrentColor());
                            }
                        }
                    }

                    // Present all automatons for the role!
                    client.automatonsHaveState(role, points, colors);

                    // Ok, so present all units that are alien to this
                    // role, but is in withing the range of atleast
                    // one of the role's observer units.
                    it = um.getAlienUnits(role).iterator();
                    while (it.hasNext()) {
                        Unit alienUnit = (Unit) it.next();
                        Iterator obsIt = um.getObservingUnits(role).iterator();
                        while (obsIt.hasNext()) {
                            Unit obsUnit = (Unit) obsIt.next();
                            UnitType uT = obsUnit.getUnitType();
                            double rangeSq = uT.getRange() * uT.getRange();
                            double dx = (alienUnit.getX() - obsUnit.getX());
                            double dy = (alienUnit.getY() - obsUnit.getY());
                            double dist = dx * dx + dy * dy;
                            if (dist <= rangeSq) {
                                // The observing unit can see the alien unit!
                                // Present alien unit to role!
                                um.activateUnit(role, alienUnit);
                                client.unitVisible(role,
						   alienUnit.getHandle(),
						   alienUnit.getName(),
						   alienUnit.getUnitType().getName(),
						   alienUnit.getX(),
						   alienUnit.getY(),
						   null,
						   null,
						   null);
                                break;
                            }
                        }
                    }
                }
            }
        }
    }


    /**
     *  Description of the Method
     *
     * @param  godsEyeRole  Description of Parameter
     */
    private void resendAllGodsEyeData(String godsEyeRole) {
        // We're already synchronized on all queues, just rock!

	UnitManager um = ReferenceHolder.getUnitManager();
	AutomatonMap am = ReferenceHolder.getAutomatonMap();

        // Make all units visible for the role!
        Iterator it = um.getAllUnits().iterator();
        while (it.hasNext()) {
            Unit unit = (Unit) it.next();
            // Present unit as if the gods eye role was an observer!
            if (!unit.isDestroyed()) {
                um.activateUnit(godsEyeRole, unit);
                client.unitVisible(godsEyeRole,
				   unit.getHandle(),
				   unit.getName(),
				   unit.getUnitType().getName(),
				   unit.getX(),
				   unit.getY(),
				   null,
				   null,
				   unit.getProperties());
            }
        }

        // Present all automatons!
        LinkedList points = new LinkedList();
        LinkedList colors = new LinkedList();
        for (int y = 0; y < am.getHeight(); y++) {
            for (int x = 0; x < am.getWidth(); x++) {
                points.add(new Point(x, y));
                colors.add(am.getAutomaton(x, y).getCurrentColor());
            }
        }
        client.automatonsHaveState(godsEyeRole, points, colors);
    }


    /**
     *  This method should be called whenever an action becomes
     *  blocked or unblocked.
     *
     * @param  unit       The unit whose action gets blocked or unblocked.
     * @param  action     The action that is now blocked or unblocked.
     * @param isBlocked true if the action is blocked, false if it is unblocked.
     */
    public void actionBlocked(Unit unit, ActionDefinition action,
			      boolean isBlocked) {

	// Tell the unit's commander that the action is now (un)blocked.
	client.unitActionBlocked(unit.getCommander(), unit.getHandle(),
				 action, isBlocked);
    }


    /**
     *  This method is to be called whenever the execution of an
     *  action is finished.
     *
     * @param  action         The action.
     * @param wasSuccessful   true if the action was successful, false
     *                             if it failed.
     */
    public void actionCompleted(Action action, boolean wasSuccessful) {
	String commander = action.getUnit().getCommander();
	int actionHandle = action.getActionHandle();

	// Tell the unit's commander whether or not the action was
	// successful.
	if (wasSuccessful) {
	    client.actionCompleted(commander, actionHandle);
	} else {
	    client.actionFailed(commander, actionHandle);
	}
    }


    /**
     *  This method should be called whenever a unit property changes.
     *
     * @param  unit      The unit whose property changed.
     * @param  key       The property that has changed.
     * @param  newValue  The new value of the property.
     */
    public void unitPropertyChanged(Unit unit, String key,
				    String newValue) {
	int handle = unit.getHandle();

	// Tell all observers (and all God's Eyes) that the property has changed.
	Iterator observers = unit.getObservers().iterator();
	while (observers.hasNext()) {
	    client.unitPropertyChanged((String) observers.next(),
				       handle, key, newValue);
	}

	// Tell all Gods Eyes too.
	Iterator godsEyes = ReferenceHolder.getUnitManager().getGodsEyeRoles().iterator();
	while (godsEyes.hasNext()) {
	    client.unitPropertyChanged((String) godsEyes.next(),
				       handle, key, newValue);
	}
    }


    /**
     *  This method is to be called when a role requests that an
     *  action be carried out, and that is for some reason impossible.
     *
     * @param  role           The role.
     * @param  actionHandle   The action handle.
     */
    public void actionRejected(String role, int actionHandle) {
	client.actionFailed(role, actionHandle);
    }


    /**
     *  This method may be called to pass some information to a
     *  specific role redarding a unit. If message is null, no message
     *  is sent.
     *
     * @param  role     The role.
     * @param  unit     The unit.
     * @param message The message to be delivered.
     */
    public void roleMessage(String role, Unit unit, String message) {

	// Make sure there is a message.
	if (message == null) {
	    return;
	}

	// Send the message.
	client.unitMessage(role, unit.getHandle(), message);
    }


    /**
     *  This method may be called to pass some information regarding a
     *  unit, such as when it has finished an action.
     *
     * @param  unit     The unit.
     * @param message The message to be delivered.
     */
    public void unitMessage(Unit unit, String message) {

	// Make sure there is a message.
	if (message == null) {
	    return;
	}

	// Obtain the unit handle.
        int unitHandle = unit.getHandle();

        // Send to all observers!
        Iterator it = unit.getObservers().iterator();
        while (it.hasNext()) {
            String role = (String) it.next();
            client.unitMessage(role, unitHandle, message);
        }

        // Send to all God's Eye roles!
        it = ReferenceHolder.getUnitManager().getGodsEyeRoles().iterator();
        while (it.hasNext()) {
            String role = (String) it.next();
            client.unitMessage(role, unitHandle, message);
        }
    }


    /**
     *  This method may be called to tell the clients that some time
     *  has passed.
     *
     * @param time  A String representing the current game time.
     */
    public void timeUpdate(String time) {
	client.timeUpdate(time);
    }

}
