/***********************************************************************

GECCO, a platform for command and control gaming

Copyright (C) 2001 Henrik Brnhielm, Andreas Enblom, Niklas
Hallenfur, Karl Hasselstrm, Henrik Hgerstrm, Oskar Linde, Jing Fu
Zi and Jon slund

This program is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 2 of the License, or (at
your option) any later version.

This program is distributed in the hope that it will be useful, but
WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307,
USA.

***********************************************************************/

package gecco.util;

/**
 *  This subclass of <code>Thread</code> has been extended with signal
 *  facilities. Other threads may send signals to it, requesting that it
 *  terminate or pause execution. This class only implements mechanisms for
 *  supporting signals; it is up to the <code>run()</code> method to check
 *  regularly whether a signal has been received, and act upon it.
 *
 * @author     Karl Hasselstrm
 * @created    July 17, 2001
 */
public class SignalThread extends Thread implements SignalRunnable {

    /**
     *  The currently pending signal. All access to it should be synchronized on
     *  <code>this</code> .
     */
    private int pendingSignal = SIGNONE;


    /**
     *  Constructs a new thread. Threads created this way must have overridden
     *  their <code>run()</code> method to actually do anything.
     */
    public SignalThread() {
        super();
    }


    /**
     *  Constructs a new thread.
     *
     * @param  target  The object whose <code>run()</code> method is called.
     */
    public SignalThread(SignalRunnable target) {
        super(target);
    }


    /**
     *  Constructs a new thread.
     *
     * @param  target  The object whose <code>run()</code> method is called.
     * @param  name    The name of the new thread.
     */
    public SignalThread(SignalRunnable target, String name) {
        super(target, name);
    }


    /**
     *  Constructs a new thread. Threads created this way must have overridden
     *  their <code>run()</code> method to actually do anything.
     *
     * @param  name  The name of the new thread.
     */
    public SignalThread(String name) {
        super(name);
    }


    /**
     *  Constructs a new thread.
     *
     * @param  group   The thread group.
     * @param  target  The object whose <code>run()</code> method is called.
     */
    public SignalThread(ThreadGroup group, SignalRunnable target) {
        super(group, target);
    }


    /**
     *  Constructs a new thread.
     *
     * @param  group   The thread group.
     * @param  target  The object whose <code>run()</code> method is called.
     * @param  name    The name of the new thread.
     */
    public SignalThread(ThreadGroup group, SignalRunnable target, String name) {
        super(group, target, name);
    }


    /**
     *  Constructs a new thread. Threads created this way must have overridden
     *  their <code>run()</code> method to actually do anything.
     *
     * @param  group  The thread group.
     * @param  name   The name of the new thread.
     */
    public SignalThread(ThreadGroup group, String name) {
        super(group, name);
    }


    /**
     *  This sends a signal to the thread. Signals with higher
     *  priority (lower value) take precedence over signals with lower
     *  priority (higher value) -- for example, <code>SIGQUIT</code>
     *  goes before <code>SIGRESUME</code>.
     *
     * @param signal The signal.
     */
    public void sendSignal(int signal) {

        // Note the incoming signal.
        synchronized (this) {
            pendingSignal = Math.min(pendingSignal, signal);
        }

        // Interrupt this thread if it's sleeping.
        interrupt();
    }


    /**
     *  Returns the object representing the currently executing signal thread.
     *
     * @return                         The current SignalThread.
     * @exception ClassCastException If the current thread is not a
     * SignalThread.
     */
    public static SignalThread current() throws ClassCastException {
        return (SignalThread) Thread.currentThread();
    }


    /**
     *  Returns the currently pending signal for the current thread. This is
     *  always <code>SIGNONE</code> if the thread isn't a <code>SignalThread
     *  </code>.
     *
     * @return    The currently pending signal.
     */
    public static int signalPending() {

        // Get the current thread.
        SignalThread t;
        try {
            t = SignalThread.current();
        } catch (ClassCastException e) {

            // This means it isn't a SignalThread, so return SIGNONE.
            return SIGNONE;
        }

        // Return the currently pending signal.
        synchronized (t) {
            return t.pendingSignal;
        }
    }

    
    /**
     *  This clears any pending signal for the current thread. Use
     *  this after you are finished handling a signal.
     */
    public static void clearSignal() {
	
        // Get the current thread.
        SignalThread t;
        try {
            t = SignalThread.current();
        } catch (ClassCastException e) {

            // This means it isn't a SignalThread, so we don't have to
            // do anything.
            return;
        }

	// Clear the signal.
	synchronized (t) {
	    t.pendingSignal = SIGNONE;
	}

	return;
    }


    /**
     *  This utility method will cause the thread to sleep as long as the
     *  currently pending signal is <code>SIGPAUSE</code> . When that is no
     *  longer the case, it will return immediately.
     */
    public static void pauseHandler() {

	// Sleep while the current signal is SIGPAUSE.
        while (SignalThread.signalPending() == SIGPAUSE) {
            try {
                // Sleep for 1-2 minutes.
                Thread.sleep(Math.round(6e4 * (Math.random() + 1.0)));
            } catch (InterruptedException e) {
                // We were woken early, but never mind.
            }
        }

	// If the new current signal is SIGRESUME, eat it.
	if (SignalThread.signalPending() == SIGRESUME) {
	    SignalThread.clearSignal();
	}
    }

}
