/***********************************************************************

GECCO, a platform for command and control gaming

Copyright (C) 2001 Karl Hasselstrm

This program is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 2 of the License, or (at
your option) any later version.

This program is distributed in the hope that it will be useful, but
WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307,
USA.

***********************************************************************/

package gecco.util;

import java.util.*;

/**
 *  A utility class for holding, parsing and comparing version
 *  numbers.
 *
 * @author     Karl Hasselstrm
 * @created July 6, 2001
 */
public class VersionNumber implements Comparable {

    /**
     *  The Vector used to hold the Integers that represent the parts
     *  of the version number.
     */
    private Vector numbers;


    /**
     *  Constructs an empty version number.
     */
    public VersionNumber() {
        numbers = new Vector();
    }


    /**
     *  Constructs a version number with only a major number, such as
     *  "3".
     *
     * @param major The major version number.
     */
    public VersionNumber(int major) {
        this();
        numbers.add(new Integer(major));
    }


    /**
     *  Constructs a version number with a major and a minor number,
     *  such as "3.1".
     *
     * @param  major  The major version number.
     * @param minor The minor version number.
     */
    public VersionNumber(int major, int minor) {
        this(major);
        numbers.add(new Integer(minor));
    }


    /**
     *  Constructs a version number with a major and two minor
     *  numbers, such as "3.1.41".
     *
     * @param major The major version number.
     * @param minor1 The most significant minor version number.
     * @param minor2 The least significant minor version number.
     */
    public VersionNumber(int major, int minor1, int minor2) {
        this(major, minor1);
        numbers.add(new Integer(minor2));
    }


    /**
     *  Constructs a version number from a String, such as
     *  "3.14.15". The String consists of a sequence of integers,
     *  separated by periods.
     *
     * @param version The String to be parsed.
     * @exception NumberFormatException If the parts of the String
     * separated by periods could not be parsed as integers.
     */
    public VersionNumber(String version) throws NumberFormatException {
        this();
        StringTokenizer tok = new StringTokenizer(version, ".");
        while (tok.hasMoreTokens()) {
            numbers.add(Integer.valueOf(tok.nextToken()));
        }
    }


    /**
     *  Compares this object with another object. Will throw a
     *  ClassCastException if the other object is not a VersionNumber.
     *
     * @param  o  The other object.
     * @return 0 if the two objects are equal, negative if this object
     * is less than the other object, and positive if this object is
     * greater than the other object.
     */
    public int compareTo(Object o) {
	try {
	    return this.compareTo((VersionNumber) o);
	} catch (ClassCastException e) {
	    throw new ClassCastException("Cannot compare a VersonNumber " +
					 "with an object of a different class.");
	}
    }

    /**
     *  Compares this VersionNumber with another VersionNumber.
     *
     * @param  other  The other VersionNumber.
     * @return 0 if the two VersionNumbers are equal, negative if this
     * VersionNumber is less than the other VersionNumber, and
     * positive if this VersionNumber is greater than the other
     * VersionNumber.
     */
    public int compareTo(VersionNumber other) {
        int comparison = 0;
        Iterator i = this.numbers.iterator();
        Iterator j = other.numbers.iterator();

        // Compare first major numbers, then minor, and so on,
        // stopping as soon as a difference is found.
        while (i.hasNext() && j.hasNext()) {
            comparison = ((Integer) i.next()).compareTo(j.next());
            if (comparison != 0) {
                return comparison;
            }
        }

        // If every corresponding element is equal, the shortest
        // string is less. If they are equally long, it's a tie.
        return this.numbers.size() - other.numbers.size();
    }


    /**
     *  Determines if this VersionNumber is less than the other
     *  VersionNumber.
     *
     * @param  other  The other VersionNumber.
     * @return true if this VersionNumber is less than the other
     *  VersionNumber, false otherwise.
     */
    public boolean lessThan(VersionNumber other) {
        return (this.compareTo(other) < 0);
    }


    /**
     *  Determines if this VersionNumber is greater than the other
     *  VersionNumber.
     *
     * @param  other  The other VersionNumber.
     * @return true if this VersionNumber is greater than the other
     * VersionNumber, false otherwise.
     */
    public boolean greaterThan(VersionNumber other) {
        return (this.compareTo(other) > 0);
    }


    /**
     *  Determines if this VersionNumber is greater than or equal to
     *  the other VersionNumber.
     *
     * @param  other  The other VersionNumber.
     * @return true if this VersionNumber is greater than or equal to
     * the other VersionNumber, false otherwise.
     */
    public boolean atLeast(VersionNumber other) {
        return (this.compareTo(other) >= 0);
    }


    /**
     *  Determines whether this object is equal to another object.
     *
     * @param  o  The other object.
     * @return    true if the objects are equal, false otherwise.
     */
    public boolean equals(Object o) {
	if (o instanceof VersionNumber) {
	    return (compareTo(o) == 0);
	} else {
	    return false;
	}
    }


    /**
     *  Computes the hash code of this VersionNumber.
     *
     * @return    An int suitable for use as a hash code.
     */
    public int hashCode() {
        return numbers.hashCode();
    }


    /**
     *  Returns the String representation of this VersionNumber.
     *
     * @return    The String representation of this VersionNumber.
     */
    public String toString() {
        StringBuffer buf = new StringBuffer();
        if (numbers.size() > 0) {
            buf.append(numbers.get(0));
            for (int i = 1; i < numbers.size(); i++) {
                buf.append(".");
                buf.append(numbers.get(i));
            }
        }
        return buf.toString();
    }
}
